"use strict";
/* Copyright © 2019 Ganchrow Scientific, SA all rights reserved */
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * The AsyncLazyMap class provides a Map which is able to create a default value
 * asynchronously.if no value is present it will return a promise immediately when the get method is call.
 * @class LazyMap
 * @example Example usage:
 * let lm = new AsyncLazyMap<string, string>((map, key) => {
 *  return Promise.resolve('defaultValue ' + key);
 * });
 * lm.set('key1', 'value');
 * await lm.ensure('key1')
 * // returns 'value'
 * await lm.ensure('key2')
 * // returns 'defaultValue key2'
 * // You can also provide the defaultValueFactory for a particular get call.
 * await lm.ensure('key2', (map, key) => {
 *  return Promise.resolve('other ' + key);
 * })
 * // returns 'other key2'
*/
class AsyncLazyMap extends Map {
    constructor(defaultValueFactory) {
        super();
        this.defaultValueFactory = defaultValueFactory;
    }
    async ensure(key, defaultValueFactory = this.defaultValueFactory) {
        if (!this.has(key)) {
            this.set(key, await defaultValueFactory(this, key));
        }
        return this.get(key);
    }
}
exports.AsyncLazyMap = AsyncLazyMap;
//# sourceMappingURL=asyncLazyMap.js.map