/* Copyright © 2016-2023 Ganchrow Scientific, SA all rights reserved */
'use strict';
Object.defineProperty(exports, "__esModule", { value: true });
const yaml = require("js-yaml");
const path = require("path");
const fs = require("fs");
const gsLogger_1 = require("./gsLogger");
const yamlExtensions_1 = require("./yamlExtensions");
const EXECUTION_ENVIRONMENT = 'EXECUTION_ENVIRONMENT';
const ENVIRONMENTS = 'ENVIRONMENTS';
const DEFAULT_EXECUTION_ENVIRONMENT = 'DEVELOPMENT';
const STRICT_ENVIRONMENT_MODE = 'strict_environment_mode';
const logger = gsLogger_1.getLogger('config-loader');
const INCLUDE_PATTERN = /^#\s+include\s+(.*)\n/;
class ConfigLoaderError extends Error {
    constructor(msg) {
        super(`ConfigLoader: ${msg}`);
    }
}
class ConfigLoader {
    constructor(executionEnvironment, basePath = '') {
        this.executionEnvironment = executionEnvironment;
        this.basePath = basePath;
        this.executionEnvironment = this.getExecutionEnvironmentFromEnv(executionEnvironment);
        if (basePath && !basePath.endsWith(path.sep)) {
            basePath += path.sep;
        }
        else {
            basePath = '';
        }
        logger.debug(`ConfigLoader EXECUTION_ENVIRONMENT: ${this.executionEnvironment} basePath ${this.basePath}`);
    }
    static viewEnvironments(fileName) {
        return Object.keys(ConfigLoader.loadConfig(fileName)[ENVIRONMENTS] || {});
    }
    static viewEnvironmentsRaw(rawYaml) {
        return Object.keys(ConfigLoader.loadConfigRaw(rawYaml)[ENVIRONMENTS] || {});
    }
    static loadConfig(fileName, baseName = '') {
        let yamlString = fs.readFileSync(fileName, 'utf-8');
        if (INCLUDE_PATTERN.test(yamlString)) {
            let baseYamlStringList = RegExp.$1.split(',').map(file => {
                return fs.readFileSync(path.resolve(baseName, file.trim()), 'utf-8');
            });
            yamlString = baseYamlStringList.concat(yamlString).join('\n');
        }
        return ConfigLoader.loadConfigRaw(yamlString, baseName);
    }
    static loadConfigRaw(yamlString, baseName = '') {
        return yaml.safeLoad(yamlString, { schema: yamlExtensions_1.schemaFactory(baseName) });
    }
    loadConfig(fileName) {
        let parsedYaml = ConfigLoader.loadConfig(`${this.basePath}${fileName}`, path.dirname(`${this.basePath}${fileName}`));
        return this.applyEnvironment(parsedYaml);
    }
    loadConfigRaw(yamlString, baseName = '') {
        let parsedYaml = ConfigLoader.loadConfigRaw(yamlString, baseName);
        return this.applyEnvironment(parsedYaml);
    }
    applyEnvironment(config = {}) {
        let allEnvironments = config[ENVIRONMENTS] || {};
        let thisEnvironment = this.matchEnvironments(allEnvironments, config[STRICT_ENVIRONMENT_MODE]);
        delete config[ENVIRONMENTS];
        Object.assign(config, thisEnvironment);
        return config;
    }
    getExecutionEnvironmentFromEnv(initialExecutionEnvironment) {
        let executionEnvironment;
        let envString;
        if (initialExecutionEnvironment) {
            executionEnvironment = initialExecutionEnvironment;
            envString = String(executionEnvironment);
        }
        else {
            envString = process.env[EXECUTION_ENVIRONMENT];
            envString = envString ? envString.toUpperCase() : DEFAULT_EXECUTION_ENVIRONMENT;
        }
        return envString;
    }
    matchEnvironments(allEnvironments, strictEnvironmentMode = false) {
        if (allEnvironments[this.executionEnvironment]) {
            return allEnvironments[this.executionEnvironment];
        }
        let possibleEnvironmentKeys = Object.keys(allEnvironments).filter(e => /\*$/.test(e));
        let curEnv = possibleEnvironmentKeys.find(env => this.generateEnvironmentExpression(env).test(this.executionEnvironment));
        let config = allEnvironments[curEnv];
        if (!config && strictEnvironmentMode) {
            throw new ConfigLoaderError(`Environment ${this.executionEnvironment} not defined`);
        }
        return config || {};
    }
    generateEnvironmentExpression(env) {
        return new RegExp(env.replace(/[.]?\*$/, '.*'));
    }
}
exports.ConfigLoader = ConfigLoader;
let defaultConfigLoader = new ConfigLoader();
function loadConfig(fileName) {
    return defaultConfigLoader.loadConfig(fileName);
}
exports.loadConfig = loadConfig;
function loadConfigRaw(yamlString, baseName = '') {
    return defaultConfigLoader.loadConfigRaw(yamlString, baseName);
}
exports.loadConfigRaw = loadConfigRaw;
//# sourceMappingURL=configLoader.js.map