"use strict";
/* Copyright © 2019 Ganchrow Scientific, SA all rights reserved */
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * The LazyMap class provides a Map which is able to create a default value
 * if no value is present and returns it immediately when the get method is call.
 * @class LazyMap
 * @example Example usage:
 * let lm = new LazyMap<string, string>((map, key) => {
 *  return 'defaultValue ' + key;
 * });
 * lm.set('key1', 'value');
 * lm.ensure('key1')
 * // returns 'value'
 * lm.ensure('key2')
 * // returns 'defaultValue key2'
 * // You can also provide the defaultValueFactory for a particular get call.
 * lm.ensure('key2', (map, key) => {
 *  return 'other ' + key;
 * })
 * // returns 'other key2'
*/
class LazyMap extends Map {
    constructor(defaultValueFactory) {
        super();
        this.defaultValueFactory = defaultValueFactory;
    }
    ensure(key, defaultValueFactory = this.defaultValueFactory) {
        if (!this.has(key)) {
            this.set(key, defaultValueFactory(this, key));
        }
        return this.get(key);
    }
}
exports.LazyMap = LazyMap;
//# sourceMappingURL=lazyMap.js.map