/* Copyright © 2020-2024 Ganchrow Scientific, SA all rights reserved */
'use strict';
Object.defineProperty(exports, "__esModule", { value: true });
const delay_1 = require("./delay");
const indexQueue_1 = require("./indexQueue");
const events_1 = require("events");
class TaskQueue {
    constructor(concurrency, handler, handleError, discardTask, restoredTasks, queueDelay) {
        this.nextQueue = new indexQueue_1.IndexQueue();
        this.pendantQueue = new indexQueue_1.IndexQueue();
        this.processingTypes = Object.create(null);
        this.moreRecentTypes = Object.create(null);
        this.pausePromiseResolver = () => { };
        this.isPaused = false;
        this.working = false;
        this.isDumping = false;
        this.executingPromises = Object.create(null);
        this.sequence = 0;
        this.shouldPause = Promise.resolve();
        this.eventEmitter = new events_1.EventEmitter();
        this.maxConcurrency = concurrency;
        this.realMaxConcurrency = concurrency;
        this.handler = handler;
        this.handleError = handleError;
        this.discardTask = discardTask;
        this.queueDelay = queueDelay || 0;
        if (restoredTasks) {
            this.restore(restoredTasks);
        }
    }
    push(type, task) {
        if (this.isDumping) {
            throw 'The queue has dumped the items and is unusable';
        }
        let seq = this.sequence++;
        this.nextQueue.enqueue({ task, type, seq });
        this.moreRecentTypes[type] = seq;
        if (!this.working) {
            return this.execute();
        }
    }
    pause() {
        if (!this.isPaused) {
            this.isPaused = true;
            this.maxConcurrency = 0;
            this.shouldPause = new Promise((resolve) => { this.pausePromiseResolver = resolve; });
        }
    }
    resume() {
        if (this.isPaused && !this.isDumping) {
            this.isPaused = false;
            this.maxConcurrency = this.realMaxConcurrency;
            this.pausePromiseResolver();
        }
    }
    async dump() {
        if (!this.isDumping) {
            this.isDumping = true;
            let dumpedTasks = Object.create(null);
            this.pause();
            // we need to give time to the handle function to put all the
            // tasks that couldn't be catched by the pause into the executingPromises object.
            await delay_1.delay(0);
            let executingPromisesArray = Object.values(this.executingPromises);
            await Promise.all(executingPromisesArray);
            while (!this.nextQueue.isEmpty()) {
                let item = this.nextQueue.dequeue();
                dumpedTasks[item.seq] = item;
            }
            while (!this.pendantQueue.isEmpty()) {
                let item = this.pendantQueue.dequeue();
                dumpedTasks[item.seq] = item;
            }
            this.eventEmitter.emit('dump', dumpedTasks);
        }
    }
    onDump(listener) {
        this.eventEmitter.addListener('dump', listener);
    }
    isProccesingType(type) {
        return this.processingTypes[type];
    }
    restore(tasks) {
        Object.values(tasks)
            .sort((a, b) => a.seq - b.seq)
            .forEach(taskItem => {
            this.push(taskItem.type, taskItem.task);
        });
    }
    pushBackFactory(type, task, seq) {
        return () => {
            this.pendantQueue.enqueue({ task, type, seq });
            if (!this.working) {
                return this.execute();
            }
        };
    }
    async handle({ type, task, seq }) {
        if (this.moreRecentTypes[type] === seq) {
            this.processingTypes[type] = true;
            let worker = Promise.resolve()
                .then(() => this.handler(type, task))
                .catch((e) => {
                if (typeof this.handleError === 'function') {
                    this.handleError(e, type, task, this.pushBackFactory(type, task, seq));
                }
            })
                .finally(() => {
                this.processingTypes[type] = false;
                delete this.executingPromises[seq];
            });
            this.executingPromises[seq] = worker;
            let executingPromisesArray = Object.values(this.executingPromises);
            if (executingPromisesArray.length >= this.maxConcurrency) {
                // here is when the magic happens, when the maxConcurrency is got, you wait
                // until any of the promises is resolved.
                // this way you stop the queue processing until you are ready to go on.
                await Promise.race(executingPromisesArray);
            }
        }
        else {
            if (typeof this.discardTask === 'function') {
                this.discardTask(type, task);
            }
        }
    }
    async execute() {
        this.working = true;
        while (!this.nextQueue.isEmpty() || !this.pendantQueue.isEmpty()) {
            await this.shouldPause;
            if (!this.pendantQueue.isEmpty()) {
                let pendant = this.pendantQueue.dequeue();
                if (this.isProccesingType(pendant.type)) {
                    if (this.moreRecentTypes[pendant.type] === pendant.seq) {
                        this.pendantQueue.enqueue(pendant);
                    }
                    else {
                        if (typeof this.discardTask === 'function') {
                            this.discardTask(pendant.type, pendant.task);
                        }
                    }
                }
                else {
                    await this.handle(pendant);
                }
            }
            if (!this.nextQueue.isEmpty()) {
                let next = this.nextQueue.dequeue();
                if (this.isProccesingType(next.type)) {
                    this.pendantQueue.enqueue(next);
                }
                else {
                    await this.handle(next);
                }
            }
            // this one is here to avoid blocking the eventloop.
            await delay_1.delay(this.queueDelay);
        }
        this.working = false;
    }
}
exports.TaskQueue = TaskQueue;
//# sourceMappingURL=taskQueue.js.map