/* Copyright © 2017-2021 Ganchrow Scientific, SA all rights reserved */
'use strict';
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const fs = require("fs");
const js_yaml_1 = require("js-yaml");
const range_1 = require("./range");
const utilities_1 = require("./utilities");
class RandomElement extends js_yaml_1.Type {
    constructor() {
        super('!randomElement', {
            kind: 'sequence',
            construct(array) {
                return utilities_1.shuffleArray(array)[0];
            },
            resolve(array) {
                return Array.isArray(array);
            }
        });
    }
}
/**
 * A custom yaml type that concatenates strings
 * @type {Concat}
 */
class Concat extends js_yaml_1.Type {
    constructor() {
        super('!concat', {
            kind: 'sequence',
            construct(items) {
                return items.join('');
            },
            resolve(items) {
                return Array.isArray(items);
            }
        });
    }
}
/**
 * A custom yaml type that loads an environment variable
 * @type {LoadfromEnv}
 */
class LoadFromEnv extends js_yaml_1.Type {
    constructor() {
        super('!loadFromEnv', {
            kind: 'sequence',
            construct(items) {
                var _a;
                const [varName, fallback] = items;
                if (process.env[varName] === undefined && fallback === undefined) {
                    throw Error(`${varName} does not exist and no fallback has been specified`);
                }
                return _a = process.env[varName], (_a !== null && _a !== void 0 ? _a : fallback);
            },
            resolve(items) {
                return Array.isArray(items);
            }
        });
    }
}
/**
 * A custom yaml type that populates yaml field with items from relative file
 * @type {FromFile}
 */
class FromFile extends js_yaml_1.Type {
    constructor(basePath) {
        super('!fromFile', {
            kind: 'sequence',
            construct(fileAndDefault) {
                let file = fileAndDefault[0];
                try {
                    return fs.readFileSync(/^\//.test(file) ? file : basePath + path.sep + file).toString('utf8').trim().split('\n');
                }
                catch (e) {
                    return fileAndDefault[1];
                }
            },
            resolve(fileAndDefault) {
                return fileAndDefault.length === 2;
            }
        });
    }
}
/**
 * A custom yaml type that converts a range into an array
 * @type {Range}
 */
class Range extends js_yaml_1.Type {
    constructor() {
        super('!range', {
            kind: 'sequence',
            construct: (data) => {
                return Array.isArray(data) && data.length === 1 ? this.transform(data[0]) : [];
            }
        });
    }
    transform(v) {
        v = String(v);
        if (/^-*[0-9.]+\.\.\.-*[0-9.]+/.test(v)) {
            return new range_1.RangeExclusive(v);
        }
        if (/^-*[.0-9]+\.\.-*[0-9.]+/.test(v)) {
            return new range_1.RangeInclusive(v);
        }
        return [];
    }
}
/**
 * A custom yaml type that describes a path relative to the file being loaded
 * @type {Path}
 */
class Path extends js_yaml_1.Type {
    constructor(basePath) {
        super('!path', {
            kind: 'scalar',
            construct(data) {
                return basePath + path.sep + data;
            },
            resolve(data) {
                return typeof data === 'string';
            }
        });
    }
}
/**
 * A custom yaml type that flattens yaml arrays, useful for merging in array configurations
 * @type {Flatten}
 */
class Flatten extends js_yaml_1.Type {
    constructor() {
        super('!flatten', {
            kind: 'sequence',
            construct(data) {
                return utilities_1.flattenArray(data);
            },
            resolve(data) {
                return Array.isArray(data);
            }
        });
    }
}
/**
 * A custom yaml type that expands an object into an array of length { expand: number }
 */
class Expand extends js_yaml_1.Type {
    constructor() {
        super('!expand', {
            kind: 'mapping',
            construct(data) {
                return (new Array(data.expand)).fill(0).map((_, i) => {
                    return JSON.parse(JSON.stringify(data.content).replace(/%%%/g, `${i}`));
                });
            },
            resolve(data) {
                return utilities_1.isObject(data) && utilities_1.isObject(data.content) && data.expand > 0;
            }
        });
    }
}
class Ymd extends js_yaml_1.Type {
    constructor(key = '!ymd') {
        super(key, {
            kind: 'scalar',
            construct(data) {
                let match = data.match(Ymd.PATTERN) || [];
                let step = match[1] === 'last' ? -1 : 1;
                let day = match[2] || 0;
                let newDate = new Date(Date.now() + step * Ymd.ONE_DAY * day);
                return this.dateFormat(newDate);
            }
        });
    }
}
Ymd.PATTERN = /(last|next) ([0-9]+) day[s]?$/i;
Ymd.ONE_DAY = 1000 * 60 * 60 * 24;
class YmdTimestamp extends Ymd {
    constructor() {
        super('!ymd_timestamp');
    }
    /* tslint:disable no-bitwise */
    dateFormat(date) {
        return (+date / 1000) | 0;
    }
}
class YmdString extends Ymd {
    dateFormat(date) {
        return `${date.toISOString().split('T')[0].replace(/\-/g, '')}`;
    }
}
function schemaFactory(basePath) {
    // See https://github.com/DefinitelyTyped/DefinitelyTyped/pull/18978
    return js_yaml_1.Schema.create(js_yaml_1.DEFAULT_SAFE_SCHEMA, [
        new Range(), new Path(basePath), new Flatten(), new YmdString(), new FromFile(basePath),
        new RandomElement(), new YmdTimestamp(), new Concat(), new Expand(), new LoadFromEnv()
    ]);
}
exports.schemaFactory = schemaFactory;
//# sourceMappingURL=yamlExtensions.js.map